import time
from functools import wraps

def retry(exceptions, total_tries=5, delay=0.5, backoff=2, silently: bool = False):
    """
    지정된 예외가 발생할 때 함수 실행을 지정된 횟수만큼 재시도하는 데코레이터입니다.

    이 데코레이터는 함수를 실행 중 예외가 발생했을 때, 주어진 조건에 따라 자동으로 함수를 재시도합니다.
    재시도 간격은 지연시간과 백오프 인자에 의해 결정되며, 모든 시도가 실패하면 최종적으로 예외를 발생시킵니다.

    Args:
        exceptions (Exception or tuple): 재시도를 트리거하는 예외 또는 예외 튜플.
        total_tries (int): 시도할 총 횟수.
        delay (float): 재시도 간 초기 지연 시간(초).
        backoff (float): 백오프 인자; 예를 들어, 2는 각 재시도 간 지연 시간을 두 배로 늘립니다.
        silently (bool): True일 경우 예외 메시지를 출력하지 않습니다.

    Returns:
        function: 예외 발생 시 재시도를 수행하는 함수.

    Examples:
        >>> @retry(Exception, total_tries=3, delay=1, backoff=1, silently=False)
        ... def test_func():
        ...     print("Trying...")
        ...     raise Exception("An error occurred")
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            current_try = 0
            while current_try < total_tries:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    current_try += 1
                    sleep_time = delay * (backoff ** (current_try - 1))
                    if current_try != total_tries:
                        if not silently:
                            print(f"{str(e)}\nRetrying in {sleep_time} seconds...")
                        time.sleep(sleep_time)
                    else:
                        print("Max retry attempts reached, aborting.")
                        raise
        return wrapper
    return decorator



from pyqqq import get_api_key
from pyqqq.utils.retry import retry
from typing import List, Union
import json
import pyqqq.config as c
import requests


class KVStore:
    """
    Simple key-value store

    Args:
        context_name (str, optional): 전략 등 DB 구분을 위한 context 식별자. 기본값은 "default".
    """

    def __init__(self, context_name: str = "default"):
        self.name = context_name

    def get(self, key: str) -> Union[str, bool, int, float, dict, list, None]:
        """
        KV store 로 부터 key 에 해당하는 값을 가져온다.

        Args:
            key (str): key

        Returns:
            str | bool | int | float | dict | list | None: value
        """
        assert type(key) == str, "key must be a string"
        assert len(key) > 0, "key must not be empty"

        url = f"{c.PYQQQ_API_URL}/kvstore/get-value"
        params = {"contextName": self.name, "key": key}
        r = self._send_request("GET", url, params=params)
        r.raise_for_status()

        if len(r.text) == 0:
            return None

        data = r.json()
        if data:
            return json.loads(data)
        else:
            return None

    def set(
        self, key: str, value: Union[str, bool, int, float, dict, list, None]
    ) -> bool:
        """
        KV store 에 key 에 해당하는 값을 저장한다.

        value는 json.dumps로 serialize 가능한 값이어야 한다.
        value가 None이면 key를 삭제한다. (delete 메소드와 동일한 효과)


        Args:
            key (str): key
            value (str | bool | int | float | dict | list | None): value

        Returns:
            bool: 성공 여부
        """
        assert type(key) == str, "key must be a string"
        assert len(key) > 0, "key must not be empty"
        assert type(value) in [
            str,
            bool,
            int,
            float,
            dict,
            list,
            None,
        ], "value must be a string, bool, int, float, dict, list, or None"

        if value is None:
            return self.delete(key)

        try:
            value = json.dumps(value)
        except:
            raise ValueError("value must be serializable to JSON")

        req_body = {
            "contextName": self.name,
            "key": key,
            "value": value,
        }

        r = self._send_request(
            "PUT", f"{c.PYQQQ_API_URL}/kvstore/set-value", json=req_body
        )
        r.raise_for_status()

        data = r.json()
        return data.get("message") == "success"

    def delete(self, key: str) -> bool:
        """
        KV store 에서 key 에 해당하는 값을 삭제한다.

        Args:
            key (str): key

        Returns:
            bool: 성공 여부
        """
        req_body = {
            "contextName": self.name,
            "key": key,
        }

        r = self._send_request(
            "DELETE", f"{c.PYQQQ_API_URL}/kvstore/delete-value", json=req_body
        )
        if r.status_code != 200:
            print(r.text)

        r.raise_for_status()

        data = r.json()
        return data.get("message") == "success"

    def keys(self) -> List[str]:
        """
        KV store 에 저장된 모든 key 를 가져온다.

        Returns:
            List[str]: key list
        """
        url = f"{c.PYQQQ_API_URL}/kvstore/list-keys"
        r = self._send_request("GET", url, params={"contextName": self.name})
        r.raise_for_status()

        data = r.json()

        return data if data else []

    def clear(self):
        """
        KV store 에 저장된 모든 key-value 를 삭제한다.
        """
        url = f"{c.PYQQQ_API_URL}/kvstore/clear"
        r = self._send_request("DELETE", url, json={"contextName": self.name})
        r.raise_for_status()

    @retry(requests.HTTPError)
    def _send_request(self, method: str, url: str, **kwargs):
        api_key = get_api_key()
        if not api_key:
            raise ValueError("API key is not set")

        return requests.request(
            method=method,
            url=url,
            headers={"Authorization": f"Bearer {api_key}"},
            **kwargs,
        )



from pyqqq import get_api_key
from pyqqq.utils.kvstore import KVStore
from tinydb import TinyDB, Query
from typing import Optional
import datetime as dtm
import pyqqq.config as c
import requests

class KISAuth:
    '''
    한국투자증권 인증 정보를 담고 있는 객체

    Args:
        appkey (str): 앱 키
        appsecret (str): 앱 시크릿
        paper_trading (bool): 모의 투자 여부
    '''
    def __init__(self, appkey: str, appsecret: str, paper_trading: bool = False):
        assert appkey is not None, "App key must be set"
        assert appsecret is not None, "App secret must be set"

        self.appkey = appkey
        self.appsecret = appsecret
        self.paper_trading = paper_trading
        self.repo = KISTokenRepository()

    @property
    def host_url(self) -> str:
        '''
        모의/실전 투자 환경에 따른 API 호스트 URL을 반환합니다.

        Returns:
            str: API 호스트 URL
        '''
        if self.paper_trading:
            return 'https://openapivts.koreainvestment.com:29443'
        else:
            return 'https://openapi.koreainvestment.com:9443'

    def get_token(self, refresh = False) -> str:
        '''
        액세스 토큰을 반환합니다.

        Args:
            refresh (bool): 강제 재발급

        Returns:
            str: 액세스 토큰
        '''
        token = self.repo.find(self.appkey, self.appsecret)
        if token is None or refresh is True:
            token, expiry = self.issue_access_token()
            self.repo.save(self.appkey, self.appsecret, token, expiry)

        return token

    def issue_access_token(self):
        '''
        액세스 토큰을 발급합니다.

        Returns:
            tuple: 액세스 토큰과 만료 시간을 담은 튜플
        '''
        if self.appkey is None or self.appsecret is None:
            raise ValueError('App key and app secret must be set')

        headers = {
            'content-type': 'application/json'
        }
        body = {
            'grant_type': 'client_credentials',
            'appkey': self.appkey,
            'appsecret': self.appsecret
        }

        r = requests.post(f"{self.host_url}/oauth2/tokenP", headers=headers, json=body)
        r.raise_for_status()

        data = r.json()

        access_token = data['access_token']
        expires_in = data['expires_in']
        expiry = int(dtm.datetime.now().timestamp() + expires_in)

        return access_token, expiry


class KISTokenRepository:
    def __init__(self):
        if get_api_key():
            self.repo = _RemoteTokenRepository()
        else:
            self.repo = _LocalTokenRepository()

    def find(self, appkey, appsecret) -> Optional[str]:
        """
        앱 키와 앱 시크릿에 해당하는 액세스 토큰을 반환합니다.

        Args:
            appkey (str): 앱 키
            appsecret (str): 앱 시크릿

        Returns:
            str|None: 액세스 토큰
        """
        return self.repo.find(appkey, appsecret)

    def save(self, appkey, appsecret, token, expiry):
        """
        앱 키와 앱 시크릿에 해당하는 액세스 토큰을 저장합니다.

        Args:
            appkey (str): 앱 키
            appsecret (str): 앱 시크릿
            token (str): 액세스 토큰
            expiry (int): 만료 시간(UNIX 타임스탬프)
        """
        return self.repo.save(appkey, appsecret, token, expiry)


class _LocalTokenRepository:
    def __init__(self):
        self.db = TinyDB(c.get_tiny_db_path())

    def find(self, appkey, appsecret) -> Optional[str]:
        TokenCache = Query()

        result = self.db.search(TokenCache.appkey == appkey and TokenCache.appsecret == appsecret and TokenCache.expiry > dtm.datetime.now().timestamp())
        if len(result) > 0:
            return result[0]['token']
        else:
            return None

    def save(self, appkey, appsecret, token, expiry):
        TokenCache = Query()
        self.db.upsert({'appkey': appkey, 'appsecret': appsecret, 'token': token, 'expiry': expiry}, TokenCache.appkey == appkey and TokenCache.appsecret == appsecret)


class _RemoteTokenRepository:
    def __init__(self):
        self.kvstore = KVStore("kis_token_cache")
        self.cache = {}

    def _cache_key(self, appkey, appsecret):
        return f"{appkey}_{appsecret}"

    def find(self, appkey, appsecret) -> Optional[str]:
        key = self._cache_key(appkey, appsecret)

        # 1st. check local data
        local_data = self.cache.get(key, None)
        if local_data:
            if local_data["expiry"] > dtm.datetime.now().timestamp():
                return local_data["token"]
            else:
                self.cache.pop(key)

        # 2nd. check remote data
        if self.kvstore.get(key):
            remote_data = self.kvstore.get(key)
            if remote_data["expiry"] > dtm.datetime.now().timestamp():
                self.cache[key] = remote_data
                return remote_data["token"]
            else:
                self.kvstore.delete(key)

        return None

    def save(self, appkey, appsecret, token, expiry):
        key = self._cache_key(appkey, appsecret)
        data = {"token": token, "expiry": expiry}
        self.cache[key] = data
        self.kvstore.set(key, data)



from typing import Dict, List, Optional
from pyqqq.utils.api_client import raise_for_status, send_request
from pyqqq.utils.logger import get_logger
import datetime as dtm
import pandas as pd
import pyqqq.config as c
import pytz

logger = get_logger("daily")


def get_all_ohlcv_for_date(date: dtm.date) -> pd.DataFrame:
    """
    주어진 날짜에 대한 모든 주식의 OHLCV(Open, High, Low, Close, Volume) 데이터를 조회합니다.

    이 함수는 특정 날짜에 대한 모든 주식의 시가, 고가, 저가, 종가 및 거래량 데이터를 API를 통해 요청하고,
    이를 pandas DataFrame 형태로 반환합니다. 반환된 DataFrame은 'code'를 인덱스로 사용합니다.

    Args:
        date (dtm.date): 조회할 날짜.

    Returns:
        pd.DataFrame: OHLCV 데이터를 포함하는 DataFrame. 'code' 컬럼은 DataFrame의 인덱스로 설정됩니다.

        DataFrame의 컬럼은 다음과 같습니다.

        - open (int): 시가.
        - high (int): 고가.
        - low (int): 저가.
        - close (int): 종가.
        - volume (int): 거래량.
        - value (int): 거래대금.
        - diff (int): 종가 대비 전일 종가의 차이.
        - diff_rate (float): 종가 대비 전일 종가의 변화율.

    Raises:
        HTTPError: API 요청이 실패했을 때 발생.

    Examples:
        >>> ohlcv_data = get_all_ohlcv_for_date(dtm.date(2023, 5, 8))
        >>> print(ohlcv_data)
                 open   high    low  close  volume       value  diff  diff_rate
        code
        000020   8710   8790   8710   8770   39019   341233350    60       0.69
        000040   1052   1133   1047   1047  590401   632158688     8       0.77
        000050   7740   7870   7700   7750    1445    11211730    10       0.13
        000070  68300  68800  67400  67600   33358  2261622200  -800      -1.17
        000075  54800  54900  54800  54800     177     9702400  -100      -0.18
    """
    url = f"{c.PYQQQ_API_URL}/domestic-stock/ohlcv/daily/all/{date}"
    r = send_request("GET", url)
    raise_for_status(r)

    data = r.json()

    cols = data["cols"]
    rows = data["rows"]

    if not rows:
        return pd.DataFrame()
    else:
        df = pd.DataFrame(rows, columns=cols)
        df.drop(columns=["date"], inplace=True)
        df.set_index("code", inplace=True)

        return df


def get_ohlcv_by_codes_for_period(
    codes: List[str], start_date: dtm.date, end_date: Optional[dtm.date] = None
) -> Dict[str, pd.DataFrame]:
    """
    지정된 코드 리스트와 기간에 대한 OHLCV 데이터를 조회합니다.

    이 함수는 하나 이상의 주식 코드와 시작 날짜, 선택적으로 종료 날짜를 지정하여 해당 기간 동안의 OHLCV 데이터를 API를 통해 요청합니다.
    반환된 데이터는 각 주식 코드별로 시가, 고가, 저가, 종가, 거래량을 포함하는 pandas DataFrame 객체들로 구성된 딕셔너리 형태로 제공됩니다.
    각 DataFrame은 해당 주식의 날짜를 기반으로 역순으로 정렬됩니다.

    Args:
        codes (List[str]): 조회할 주식 코드들의 리스트.
        start_date (dtm.date): 조회할 기간의 시작 날짜.
        end_date (Optional[dtm.date]): 조회할 기간의 종료 날짜. 지정하지 않으면 최근 거래일 까지 조회됩니다.

    Returns:
        dict: 주식 코드를 키로 하고, 해당 코드의 OHLCV 데이터를 포함하는 DataFrame을 값으로 하는 딕셔너리.

        DataFrame의 컬럼은 다음과 같습니다.

        - open (int): 시가.
        - high (int): 고가.
        - low (int): 저가.
        - close (int): 종가.
        - volume (int): 거래량.
        - value (int): 거래대금.
        - diff (int): 종가 대비 전일 종가의 차이.
        - diff_rate (float): 종가 대비 전일 종가의 변화율.

    Raises:
        HTTPError: API 요청이 실패했을 때 발생.

    Examples:
        >>> dfs = get_ohlcv_by_codes_for_period(['005930', '000660'], dtm.date(2024, 5, 7), dtm.date(2024, 5, 8))
        >>> print(dfs)
        {'319640':          date   open   high    low  close  volume      value  diff  diff_rate
         0  2024-05-08  15445  15460  15365  15435    6419   99007660    -5      -0.03
         1  2024-05-07  15355  15525  15355  15440   22318  345280470    85       0.55,
         '005930':          date   open   high    low  ...    volume          value  diff  diff_rate
         0  2024-05-08  80800  81400  80500  ...  12960682  1050108654400     0       0.00
         1  2024-05-07  79600  81300  79400  ...  26238868  2112619288066  3700       4.77
    """
    tz = pytz.timezone("Asia/Seoul")

    url = f"{c.PYQQQ_API_URL}/domestic-stock/ohlcv/daily/series"
    params = {"codes": ",".join(codes), "start_date": start_date}
    if end_date is not None:
        params["end_date"] = end_date

    r = send_request("GET", url, params=params)
    raise_for_status(r)

    data = r.json()
    cols = data["cols"]
    dataset = data["rows"]
    result = {}

    for code in dataset.keys():
        rows = dataset[code]
        for row in rows:
            dt = row[0]
            if dt[-1] == "Z":
                dt = dt[:-1] + "+00:00"
            dt = dtm.datetime.fromisoformat(dt).astimezone(tz).replace(tzinfo=None)
            row[0] = dt.date()

        rows.reverse()
        result[code] = pd.DataFrame(rows, columns=cols)

    return result



from decimal import Decimal, ROUND_CEILING, ROUND_FLOOR, ROUND_HALF_UP
from typing import Union


def quantize_krx_price(
    price: Union[Decimal, int, float], etf_etn: bool, rounding: str = "floor"
) -> int:
    """
    주어진 가격을 한국거래소(KRX)의 틱 사이즈에 따라 지정된 반올림 방식으로 조정합니다.

    이 함수는 ETF 또는 ETN 여부에 따라 적절한 틱 사이즈를 계산하고, 주어진 가격을 이 틱 사이즈에 맞추어 반올림합니다.
    사용자는 'round', 'ceil', 'floor' 중에서 반올림 방식을 선택할 수 있습니다.

    Args:
        price (Decimal): 반올림할 가격.
        etf_etn (bool): 가격이 ETF 또는 ETN 상품인 경우 True, 그 외는 False.
        rounding (str, optional): 반올림 방식('round', 'ceil', 'floor'). 기본값은 'floor'.

    Returns:
        int: 반올림된 가격.

    Raises:
        ValueError: rounding이 지정된 세 가지 옵션 중 하나가 아닐 경우 오류를 발생시킵니다.

    Examples:
        >>> quantize_krx_price(Decimal('1520.75'), False, 'round')
        1521
        >>> quantize_krx_price(Decimal('1520.75'), True, 'ceil')
        1521
        >>> quantize_krx_price(Decimal('1520.75'), False, 'floor')
        1520
    """

    if rounding not in ["round", "ceil", "floor"]:
        raise ValueError("rounding should be one of ['round', 'ceil', 'floor']")

    price = Decimal(price)

    constant_rounding = {
        "round": ROUND_HALF_UP,
        "ceil": ROUND_CEILING,
        "floor": ROUND_FLOOR,
    }[rounding]

    tick_size = get_krx_tick_size(price, etf_etn)

    return int(
        (price / tick_size).quantize(Decimal("1"), rounding=constant_rounding)
        * tick_size
    )


def get_krx_tick_size(price: float, etf_etn: bool) -> int:
    """
    주어진 가격과 금융 상품 유형에 따라 적절한 틱 사이즈를 반환합니다.

    한국거래소(KRX)의 틱 사이즈 규칙에 따라, 특정 가격대의 주식 또는 ETF/ETN의 최소 가격 변동 단위(틱 사이즈)를 결정합니다.
    입력된 price가 각 가격대의 최소값 미만일 경우 해당하는 틱 사이즈를 반환하며, 모든 조건에 부합하지 않는 경우 최대 가격을 반환합니다.

    Args:
        price (float): 상품의 가격.
        etf_etn (bool): 상품이 ETF 또는 ETN인 경우 True, 아니면 False.

    Returns:
        int: 결정된 틱 사이즈.

    Raises:
        AssertionError: price가 0 이하일 경우 오류를 발생시킵니다.

    Examples:
        >>> get_krx_tick_size(1500, False)
        1
        >>> get_krx_tick_size(2500, False)
        5
        >>> get_krx_tick_size(2500, True)
        1
    """

    assert price > 0, "price should be greater than 0"

    conds = []
    max_value = 0

    if etf_etn:
        conds = [(2000, 1)]
        max_value = 5
    else:
        conds = [
            (2000, 1),
            (5000, 5),
            (20000, 10),
            (50000, 50),
            (200000, 100),
            (500000, 500),
        ]
        max_value = 1000

    for min_price, size in conds:
        if price < min_price:
            return size

    return max_value



import os
import pyqqq.config as c
import importlib.metadata

__version__ = importlib.metadata.version("pyqqq")

_api_key = None

def set_api_key(api_key: str):
    """
    전역 PyQQQ API 키를 설정합니다.

    이 함수는 주어진 API 키를 전역 변수로 저장합니다. 입력된 API 키는 None이 아니어야 하며, 최소 32자 이상이어야 합니다.

    Args:
        api_key (str): 설정할 API 키.

    Raises:
        AssertionError: API 키가 None이거나 32자 미만일 경우 오류를 발생시킵니다.

    Examples:
        >>> set_api_key("1234567890abcdef1234567890abcdef")
    """

    global _api_key

    assert api_key is not None, "API key must not be None"
    assert len(api_key) >= 32, "API key must be at least 32 characters long"

    _api_key = api_key


def get_api_key() -> str | None:
    """
    설정된 전역 PyQQQ API 키를 반환합니다.

    API 키가 메모리에 저장되어 있으면 그 값을 반환합니다. 저장되어 있지 않다면, 환경 변수나 자격 증명 파일에서
    API 키를 조회하여 반환합니다. 모든 방법이 실패하면 None을 반환합니다.

    Returns:
        str | None: 저장된 API 키 또는 None.

    Examples:
        >>> api_key = get_api_key()
        >>> print(api_key)
    """
    if _api_key:
        return _api_key

    elif c.get_pyqqq_api_key() is not None:
        return c.get_pyqqq_api_key()

    elif os.path.exists(c.get_credential_file_path()):
        with open(c.get_credential_file_path(), "r") as f:
            return f.read().strip()

    return None
